﻿/**
 * Relative Accuracy Measures for Stroke Gestures (.NET Framework 4.5 C# version)
 *
 * 	    Radu-Daniel Vatavu, Ph.D.
 *	    University Stefan cel Mare of Suceava
 *	    Suceava 720229, Romania
 *	    vatavu@eed.usv.ro
 *
 * 	    Lisa Anthony, Ph.D.
 *      UMBC
 *      Information Systems Department
 *      1000 Hilltop Circle
 *      Baltimore, MD 21250
 *      lanthony@umbc.edu
 *	
 *	    Jacob O. Wobbrock, Ph.D.
 * 	    The Information School
 *	    University of Washington
 *	    Seattle, WA 98195-2840
 *	    wobbrock@uw.edu
 *
 * The academic publication for the relative accuracy measures, and what should be 
 * used to cite it, is:
 *
 *	Vatavu, R.-D., Anthony, L. and Wobbrock, J.O. (2013).  
 *	  Relative accuracy measures for stroke gestures. 
 *	  In Proceedings of the 15th ACM on International Conference 
 *	  on Multimodal Interaction (ICMI '13). ACM, New York, NY, USA, 279-286
 *	  DOI=10.1145/2522848.2522875 http://doi.acm.org/10.1145/2522848.2522875
 *
 * This software is distributed under the "New BSD License" agreement:
 *
 * Copyright (c) 2013, Radu-Daniel Vatavu, Lisa Anthony, and 
 * Jacob O. Wobbrock. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *    * Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    * Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *    * Neither the names of the University Stefan cel Mare of Suceava, 
 *	    University of Washington, nor UMBC, nor the names of its contributors 
 *	    may be used to endorse or promote products derived from this software 
 *	    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL Radu-Daniel Vatavu OR Lisa Anthony
 * OR Jacob O. Wobbrock BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
**/
using System;
using System.Collections.Generic;

namespace RelativeAccuracyMeasures
{
    public static class PDollarAlternatives
    {
        /// <summary>
        /// Optimal alignment of two clouds of points using the Hungarian weighted matching algorithm for bipartite graphs.
        /// Returns the matching between the two clounds.
        /// Complexity: O(n^3)
        /// </summary>
        /// <param name="g1"></param>
        /// <param name="g2"></param>
        /// <returns></returns>
        public static double Hungarian(Point[] points1, Point[] points2, ref int[] matching)
        {
            int n = points1.Length;
            double[][] weights = new double[n][];
            for (int i = 0; i < n; i++)
            {
                weights[i] = new double[n];
                for (int j = 0; j < n; j++)
                    weights[i][j] = -Measurements.SqrEuclideanDistance(points1[i], points2[j]);
            }

            matching = PerformHungarianMatching(weights);
            double cost = 0;
            for (int i = 0; i < n; i++)
                cost += -weights[i][matching[i]];
            return cost;
        }

        /// <summary>
        /// Implements the Hungarian algorithm for finding the maximum weighted matching for bipartite graphs
        /// Complexity: O(n^3)
        /// Source: Christos H. Papadimitriou, Kenneth Steiglitz. Combinatorial optimization: algorithms and complexity, Dover Publications, 1998 (p.248)
        /// </summary>
        /// <param name="weights"></param>
        /// <param name="n"></param>
        /// <returns></returns>
        private static int[] PerformHungarianMatching(double[][] weights)
        {
            // constants
            int NOT_FOUND = -1;
            int NOT_MATCHED = -1;

            // number of vertexes for the two sets of a bipartite graph
            // (the left and right sets)
            int n = weights.GetLength(0);

            // initialize vertex labeling for the left and right sets
            double[] labelsLeft = new double[n];
            double[] labelsRight = new double[n];
            for (int i = 0; i < n; i++)
            {
                labelsLeft[i] = weights[i][0];
                for (int j = 1; j < n; j++)
                    labelsLeft[i] = Math.Max(labelsLeft[i], weights[i][j]);
                labelsRight[i] = 0;
            }

            // initalize matching
            int matchingCount = 0;              // the size of the current matching (stop when reach n)
            int[] matchingLeft = new int[n];
            int[] matchingRight = new int[n];
            for (int i = 0; i < n; i++)
            {
                matchingLeft[i] = NOT_MATCHED;  // the vertex from the right set to which vertex i from the left set is matched to
                matchingRight[i] = NOT_MATCHED; // the vertex from the left set to which vertex i from the right set is matched to
            }

            bool[] visitedLeft = new bool[n];   // visitedLeft[i] = true if vertex i from the left set has been processed
            bool[] visitedRight = new bool[n];  // visitedRight[i] = true if vertex i from the right set has been processed
            int[] parent = new int[n];          // stores the edges of the Hungarian tree, parent[i] = the vertex from the left set that led to the discovery of vertex i from the right set
            while (matchingCount < n)
            {
                // pick free vertex from the left set as the root of the Hungarian tree
                int u = 0;
                while (matchingLeft[u] != NOT_MATCHED) u++;

                // update the labels until a non-matched vertex y from the right set is found
                // which will increase the cardinality of the matching
                int y = NOT_FOUND;
                while (y == NOT_FOUND)
                {
                    // reset processed vertexes
                    Array.Clear(visitedLeft, 0, n);
                    Array.Clear(visitedRight, 0, n);

                    // traverse the equality graph using the breadth-first approach
                    // we stop traversal when 
                    // 1) no more vertexes are available for exploring or 
                    // 2) we found a free vertex in the right set (denoted by y)
                    Queue<int> queue = new Queue<int>();
                    queue.Enqueue(u);
                    visitedLeft[u] = true;
                    while (queue.Count > 0 && y == NOT_FOUND)
                    {
                        int vertex = queue.Dequeue();
                        for (int j = 0; j < n; j++)
                            if (!visitedRight[j])
                            {
                                double diff = weights[vertex][j] - (labelsLeft[vertex] + labelsRight[j]);
                                if (diff < 0)
                                    diff = -diff;
                                if (diff < 10e-4f)
                                {
                                    parent[j] = vertex;
                                    visitedRight[j] = true;

                                    // check whether j from the right set is already matched to a vertex (z) from the left set
                                    int z = matchingRight[j];
                                    if (z == NOT_MATCHED)
                                    {
                                        // we found a non-matched vertex, stop traversal
                                        y = j;
                                        break;
                                    }
                                    else
                                    {
                                        // add z to the BFS queue and continue exploration
                                        queue.Enqueue(z);
                                        visitedLeft[z] = true;
                                    }
                                }
                            }
                    }

                    if (y == NOT_FOUND)
                    {
                        // update vertex lables in order to enlarge the equality graph
                        double alpha = double.MaxValue;
                        for (int i = 0; i < n; i++)
                            if (visitedLeft[i])
                                for (int j = 0; j < n; j++)
                                    if (!visitedRight[j])
                                    {
                                        double diff = labelsLeft[i] + labelsRight[j] - weights[i][j];
                                        if (alpha > diff)
                                            alpha = diff;
                                    }

                        for (int i = 0; i < n; i++)
                        {
                            if (visitedLeft[i]) labelsLeft[i] -= alpha;
                            if (visitedRight[i]) labelsRight[i] += alpha;
                        }
                    }
                    else
                    {
                        // the path from root u (left set) to y (right set) is an augmenting path
                        // and therefore we can increase the cardinality of the current matching by reversing the edges of the path
                        int index = y;
                        while (index != NOT_MATCHED)
                        {
                            int t = matchingLeft[parent[index]];
                            matchingLeft[parent[index]] = index;
                            matchingRight[index] = parent[index];
                            index = t;
                        }
                        matchingCount++;
                    }
                }
            }

            // return the optimal alignment of the two sets of vertexes
            return matchingLeft;
        }
    }
}
